@@ -25,6 +25,7 @@
 #include <string.h>
 
 #include <libxml/encoding.h>
+#include <ctype.h>
 
 #include <plist/plist.h>
 #include "plist.h"
@@ -63,6 +64,40 @@ enum
     BPLIST_MASK = 0xF0
 };
 
+static void float_byte_convert(uint8_t * address, size_t size)
+{
+#if G_BYTE_ORDER == G_LITTLE_ENDIAN && !defined (__VFP_FP__)
+    uint8_t i = 0, j = 0;
+    uint8_t tmp = 0;
+
+    for (i = 0; i < (size / 2); i++)
+    {
+        tmp = address[i];
+        j = ((size - 1) + 0) - i;
+        address[i] = address[j];
+        address[j] = tmp;
+    }
+#endif
+}
+
+union plist_uint_ptr
+{
+    void *src;
+    uint8_t *u8ptr;
+    uint16_t *u16ptr;
+    uint32_t *u32ptr;
+    uint64_t *u64ptr;
+};
+
+#define get_unaligned(ptr)			  \
+  ({                                              \
+    struct __attribute__((packed)) {		  \
+      typeof(*(ptr)) __v;			  \
+    } *__p = (void *) (ptr);			  \
+    __p->__v;					  \
+  })
+
+
 static void byte_convert(uint8_t * address, size_t size)
 {
 #if G_BYTE_ORDER == G_LITTLE_ENDIAN
@@ -79,23 +114,36 @@ static void byte_convert(uint8_t * address, size_t size)
 #endif
 }
 
-static uint32_t uint24_from_be(char *buff)
+static uint32_t uint24_from_be(union plist_uint_ptr buf)
 {
+    union plist_uint_ptr tmp;
     uint32_t ret = 0;
-    char *tmp = (char *) &ret;
-    memcpy(tmp + 1, buff, 3 * sizeof(char));
-    byte_convert(tmp, sizeof(uint32_t));
+
+    tmp.src = &ret;
+
+    memcpy(tmp.u8ptr + 1, buf.u8ptr, 3 * sizeof(char));
+
+    byte_convert(tmp.u8ptr, sizeof(uint32_t));
     return ret;
 }
 
 #define UINT_TO_HOST(x, n) \
-		(n == 8 ? GUINT64_FROM_BE( *(uint64_t *)(x) ) : \
-		(n == 4 ? GUINT32_FROM_BE( *(uint32_t *)(x) ) : \
-		(n == 3 ? uint24_from_be( (char*)x ) : \
-		(n == 2 ? GUINT16_FROM_BE( *(uint16_t *)(x) ) : \
-		*(uint8_t *)(x) ))))
+	({ \
+		union plist_uint_ptr __up; \
+		__up.src = x; \
+		(n == 8 ? GUINT64_FROM_BE( get_unaligned(__up.u64ptr) ) : \
+		(n == 4 ? GUINT32_FROM_BE( get_unaligned(__up.u32ptr) ) : \
+		(n == 3 ? uint24_from_be( __up ) : \
+		(n == 2 ? GUINT16_FROM_BE( get_unaligned(__up.u16ptr) ) : \
+		*__up.u8ptr )))); \
+	})
 
-#define be64dec(x) GUINT64_FROM_BE( *(uint64_t*)(x) )
+#define be64dec(x) \
+	({ \
+		union plist_uint_ptr __up; \
+		__up.src = x; \
+		GUINT64_FROM_BE( get_unaligned(__up.u64ptr) ); \
+	})
 
 #define get_needed_bytes(x) \
 		( ((uint64_t)x) < (1ULL << 8) ? 1 : \
@@ -136,23 +184,27 @@ static plist_t parse_real_node(char *bnode, uint8_t size)
 {
     plist_data_t data = plist_new_plist_data();
     float floatval = 0.0;
+    uint8_t* buf;
 
     size = 1 << size;			// make length less misleading
+    buf = malloc (size);
+    memcpy (buf, bnode, size);
     switch (size)
     {
     case sizeof(float):
-        floatval = *(float *) bnode;
-        byte_convert((uint8_t *) & floatval, sizeof(float));
+        float_byte_convert(buf, size);
+        floatval = *(float *) buf;
         data->realval = floatval;
         break;
     case sizeof(double):
-        data->realval = *(double *) bnode;
-        byte_convert((uint8_t *) & (data->realval), sizeof(double));
+        float_byte_convert(buf, size);
+        data->realval = *(double *) buf;
         break;
     default:
         free(data);
         return NULL;
     }
+    free (buf);
     data->type = PLIST_REAL;
     data->length = sizeof(double);
 
@@ -192,7 +244,6 @@ static plist_t parse_unicode_node(char *bnode, uint64_t size)
     uint64_t i = 0;
     gunichar2 *unicodestr = NULL;
     gchar *tmpstr = NULL;
-    int type = 0;
     glong items_read = 0;
     glong items_written = 0;
     GError *error = NULL;
@@ -627,6 +678,11 @@ static void write_int(GByteArray * bplist, uint64_t val)
     //do not write 3bytes int node
     if (size == 3)
         size++;
+
+#if G_BYTE_ORDER == G_BIG_ENDIAN
+    val = val << ((sizeof(uint64_t) - size) * 8);
+#endif
+
     buff = (uint8_t *) malloc(sizeof(uint8_t) + size);
     buff[0] = BPLIST_UINT | Log2(size);
     memcpy(buff + 1, &val, size);
@@ -637,7 +693,7 @@ static void write_int(GByteArray * bplist, uint64_t val)
 
 static void write_real(GByteArray * bplist, double val)
 {
-    uint64_t size = get_real_bytes(*((uint64_t *) & val));	//cheat to know used space
+    uint64_t size = get_real_bytes(val);	//cheat to know used space
     uint8_t *buff = (uint8_t *) malloc(sizeof(uint8_t) + size);
     buff[0] = BPLIST_REAL | Log2(size);
     if (size == sizeof(double))
@@ -649,7 +705,7 @@ static void write_real(GByteArray * bplist, double val)
         float tmpval = (float) val;
         memcpy(buff + 1, &tmpval, size);
     }
-    byte_convert(buff + 1, size);
+    float_byte_convert(buff + 1, size);
     g_byte_array_append(bplist, buff, sizeof(uint8_t) + size);
     free(buff);
 }
@@ -660,7 +716,7 @@ static void write_date(GByteArray * bplist, double val)
     uint8_t *buff = (uint8_t *) malloc(sizeof(uint8_t) + size);
     buff[0] = BPLIST_DATE | Log2(size);
     memcpy(buff + 1, &val, size);
-    byte_convert(buff + 1, size);
+    float_byte_convert(buff + 1, size);
     g_byte_array_append(bplist, buff, sizeof(uint8_t) + size);
     free(buff);
 }
@@ -677,6 +733,8 @@ static void write_raw_data(GByteArray * bplist, uint8_t mark, uint8_t * val, uin
         g_byte_array_append(bplist, int_buff->data, int_buff->len);
         g_byte_array_free(int_buff, TRUE);
     }
+    //stupid unicode buffer length
+    if (BPLIST_UNICODE==mark) size *= 2;
     buff = (uint8_t *) malloc(size);
     memcpy(buff, val, size);
     g_byte_array_append(bplist, buff, size);
@@ -702,7 +760,8 @@ static void write_unicode(GByteArray * bplist, gunichar2 * val, uint64_t size)
     memcpy(buff, val, size2);
     for (i = 0; i < size; i++)
         byte_convert(buff + i * sizeof(gunichar2), sizeof(gunichar2));
-    write_raw_data(bplist, BPLIST_STRING, buff, size2);
+    write_raw_data(bplist, BPLIST_UNICODE, buff, size);
+    free(buff);
 }
 
 static void write_array(GByteArray * bplist, GNode * node, GHashTable * ref_table, uint8_t dict_param_size)
@@ -729,6 +788,9 @@ static void write_array(GByteArray * bplist, GNode * node, GHashTable * ref_tabl
     for (i = 0, cur = node->children; cur && i < size; cur = cur->next, i++)
     {
         idx = *(uint64_t *) (g_hash_table_lookup(ref_table, cur));
+#if G_BYTE_ORDER == G_BIG_ENDIAN
+	idx = idx << ((sizeof(uint64_t) - dict_param_size) * 8);
+#endif
         memcpy(buff + i * dict_param_size, &idx, dict_param_size);
         byte_convert(buff + i * dict_param_size, dict_param_size);
     }
@@ -764,10 +826,16 @@ static void write_dict(GByteArray * bplist, GNode * node, GHashTable * ref_table
     for (i = 0, cur = node->children; cur && i < size; cur = cur->next->next, i++)
     {
         idx1 = *(uint64_t *) (g_hash_table_lookup(ref_table, cur));
+#if G_BYTE_ORDER == G_BIG_ENDIAN
+	idx1 = idx1 << ((sizeof(uint64_t) - dict_param_size) * 8);
+#endif
         memcpy(buff + i * dict_param_size, &idx1, dict_param_size);
         byte_convert(buff + i * dict_param_size, dict_param_size);
 
         idx2 = *(uint64_t *) (g_hash_table_lookup(ref_table, cur->next));
+#if G_BYTE_ORDER == G_BIG_ENDIAN
+	idx2 = idx2 << ((sizeof(uint64_t) - dict_param_size) * 8);
+#endif
         memcpy(buff + (i + size) * dict_param_size, &idx2, dict_param_size);
         byte_convert(buff + (i + size) * dict_param_size, dict_param_size);
     }
@@ -778,6 +846,20 @@ static void write_dict(GByteArray * bplist, GNode * node, GHashTable * ref_table
 
 }
 
+static int is_ascii_string(char* s, int len)
+{
+  int ret = 1, i = 0;
+  for(i = 0; i < len; i++)
+  {
+      if ( !isascii( s[i] ) )
+      {
+          ret = 0;
+          break;
+      }
+  }
+  return ret;
+}
+
 void plist_to_bin(plist_t plist, char **plist_bin, uint32_t * length)
 {
     GPtrArray *objects = NULL;
@@ -858,17 +940,16 @@ void plist_to_bin(plist_t plist, char **plist_bin, uint32_t * length)
         case PLIST_KEY:
         case PLIST_STRING:
             len = strlen(data->strval);
-            type = xmlDetectCharEncoding(data->strval, len);
-            if (XML_CHAR_ENCODING_UTF8 == type)
+            if ( is_ascii_string(data->strval, len) )
+            {
+                write_string(bplist_buff, data->strval);
+            }
+            else
             {
                 unicodestr = g_utf8_to_utf16(data->strval, len, &items_read, &items_written, &error);
                 write_unicode(bplist_buff, unicodestr, items_written);
                 g_free(unicodestr);
             }
-            else if (XML_CHAR_ENCODING_ASCII == type || XML_CHAR_ENCODING_NONE == type)
-            {
-                write_string(bplist_buff, data->strval);
-            }
             break;
         case PLIST_DATA:
             write_data(bplist_buff, data->buff, data->length);
@@ -897,7 +978,12 @@ void plist_to_bin(plist_t plist, char **plist_bin, uint32_t * length)
     for (i = 0; i < num_objects; i++)
     {
         uint8_t *offsetbuff = (uint8_t *) malloc(offset_size);
-        memcpy(offsetbuff, offsets + i, offset_size);
+
+#if G_BYTE_ORDER == G_BIG_ENDIAN
+	offsets[i] = offsets[i] << ((sizeof(uint64_t) - offset_size) * 8);
+#endif
+
+        memcpy(offsetbuff, &offsets[i], offset_size);
         byte_convert(offsetbuff, offset_size);
         g_byte_array_append(bplist_buff, offsetbuff, offset_size);
         free(offsetbuff);
